@function foo() {
	$limit: 10;
	$y: 0;
	@for $x from 1 through $limit {
	  $limit: 4;
	  $y: $y + $x;
	}
	@return $y;
}

@function bar() {
  $x: 0;
  @for $i from 1 through 10 {
    $x: $x + $i;
  }
  @return $x;
}

div {
	width: foo();
  height: bar();
}