@function foo($c) {
  @return hue($c);
}

@mixin moo($c) {
  border: 1px solid foo($c);
  @include booboo;
}

@mixin moomoo($c) {
  @include moo($c);
  color: gray;
}

div {
  @include moomoo(fudge);
}

// div by zero
// called from foo line 2
// called from moo line 6
// called from moomoo line 10
// called from line 15