div {
  hoo: 3/4;
  goo: 3/4;
  foo: 15/3/4;
  foo: 1;
  coo: 2/3/3/4;
  bar: true;
  bar: false and true;
  bar: false;
  blah: a;
  blah: 0.75;
  blah: b;
  bloo: 3/4;
  blix: "hey 3/4 ho"; }

@media screen and (hux: 3 / 4) {
  div {
    color: red; } }
div {
  blah: "ho 3/4 ho"; }

span {
  fludge: 3/4; }
